# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>

bl_info = {
    "name": "Export UV Layout to SVG",
    "author": "Witold Jaworski, Campbell Barton, Matt Ebb",
    "version": (1, 1, 1),
    "blender": (2, 6, 6),
    "location": "Image-Window > UVs > Export to SVG",
    "description": "Export the UV layout to an SVG (vector) file",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Import-Export"}

#######################################################################################
# This code is based on the Export UV Layout" Python package, which is the
# part of the standard Blender (i.e. the "official" BF add-on).
# It was written by Campbell Barton and Matt Ebb, and can handle several
# different output formats (SVG, PNG, EPS).
# This add-on is a "specialized" exporter that stores the data in the SVG format, only.
# I have written it, because I had missed in the standard one some features important
# for my workflow. The standard "Export UV Layout" add-on package contains four Python
# modules (the main one and one per each supported format). This "quick and dirty" 
# adaptation contains "glued" code from two of them (__init__.py and export_uv_svg.py).
# That's why its code is in some places more complicated than it should be. I have 
# preserved many part of the original code because it just works.  
#######################################################################################
#--- Update history
# 2013-03-02 (Witold Jaworski): adapted to Blender 2.66 (Mesh.uv_loop_layers => Mesh.uv_layers)

#---#### Some constants
_DEBUG = 0 #Set it to 1 or 2 to see some diagnostic messages on the Blender system console
THIS_ADDON = 'io_mesh_uv_svg.wjaworski.pl' #unique add-on name (to store add-on internal data in the scene)

#---#### Imports 
import bpy

from bpy.props import (StringProperty,
                       BoolProperty,
                       IntVectorProperty,
                       FloatVectorProperty,
                       FloatProperty,
                       )
from bpy.utils import register_module, unregister_module
from bpy.app import version_string
from bpy.path import ensure_ext, clean_name
from xml.sax.saxutils import escape
from os.path import basename, splitext, split, join

#---##### Blender API interface
class ExportUVLayoutToSVG(bpy.types.Operator):
    """Export UV layout to a SVG file"""

    bl_idname = "uv.export_svg_layout"
    bl_label = "Export to SVG"
    # bl_options = {'REGISTER', 'UNDO'}

    #---### Mandatory UI properties (required by the File Browser window)
    filepath = StringProperty(
            subtype='FILE_PATH',
            ) #Placed in the file path field of the File Browser window
    
    check_existing = BoolProperty(
            name="Check Existing",
            description="Check and warn on overwriting existing files",
            default=True,
            options={'HIDDEN'},
            ) #Required to mark the potential file overwritting, in red 
    
    #---### Additional UI properties (exposed in the helper File Browser panel)
    export_all = BoolProperty(
            name="All UVs",
            description="Export all UVs in this mesh (not just visible ones)",
            default=False,
            ) 

    export_selected = BoolProperty(
            name="From All Objects",
            description="Include in the export UVs of all selected objects (not only the active one)",
            default=False,
            )
    modified = BoolProperty(
            name="Use Object Modifiers",
            description="Temporarily apply the object modifiers before exporting the mesh UVs",
            default=False,
            )
    size = IntVectorProperty(
            size=2,
            default=(1024, 1024),
            min=8, max=32768,
            name = "Image Size",
            description="Dimensions of the exported file",
            )
    stroke_color = FloatVectorProperty(
            name="Edge Color",
            description="Color of the SVG face outline",
            subtype='COLOR',
            size = 3,
            min=0.0, max=1.0,
            default=(0.0,0.0,0.0)
            )
    stroke_width = FloatProperty(
            name="Edge Width",
            description="Width of the SVG face outline (px)",
            min=0.0, soft_max=10,
            default=0.5
            )
    
    override_color = BoolProperty(
            name="Override Fill Color:",
            description="Use the specified color to fill the SVG faces, instead of the material diffuse/display color",
            default=False
            )
    fill_color = FloatVectorProperty(
            name="Color",
            description="Color for the SVG faces",
            subtype='COLOR',
            size = 3,
            min=0.0, max=1.0,
            default=(1.0,1.0,1.0)
            )
    opacity = FloatProperty(
            name="Fill Opacity",
            description="Opacity of the SVG faces",
            min=0.0, max=1.0,
            default=0.25
            )
    use_object_name = BoolProperty(
            name="Use Object Name",
            description="Use the name of the current object (and its UV Map) instead of the specified filename",
            default=False
            )
    add_border = BoolProperty(
            name="Border Around Image",
            description="Draw a border around the SVG image bounds",
            default=False
            )
    
    #---### Other non-UI attributes
    mode = "SVG" #in the original script it was an UI property (a dropdown list of supported formats)

    #attributes from this list are preserved in the scene data:
    _stored_attributes = ['filepath', 
                          'export_all', 'export_selected',
                          'modified', 'size', 
                          'stroke_color', 'stroke_width', 
                          'override_color', 'fill_color',
                          'opacity', 'use_object_name', 'add_border']

    #---### Helper internal functions
    def _space_image(self, context):
        """Function returns an object that represents the UV/Image Editor, or None 
            Arguments:
            @context (Context): context received from Blender caller
        """
        space_data = context.space_data
        if isinstance(space_data, bpy.types.SpaceImageEditor):
            return space_data
        else:
            return None

    def _image_size(self, context, default_width=1024, default_height=1024):
        """Function returns image size (as a two-element tuple)  
            Arguments:
            @context (Context): context received from Blender caller
            @default_width (int): default result value
            @default_height (int): default result value
            
            Default values are returned in the result tuple when there is no image
        """
        image_width, image_height = default_width, default_height

        space_data = self._space_image(context)
        if space_data:
            image = space_data.image
            if image:
                width, height = tuple(context.space_data.image.size)
                # when the data is valid:
                if width and height:
                    image_width, image_height = width, height

        return image_width, image_height
    
    def _default_filepath(self):
        """Function returns the path and name of the current Blender file, but without any extension  
        """
        return splitext(bpy.data.filepath)[0]
    
    def _object_filename(self, obj):
        """Function returns the path and name derived from the given object  
            Arguments:
            @obj (Object): object, which name (and eventual UV Map name) will form the file name
            
            (It is used when the option <Use Object Name> is enabled
        """
        name = clean_name(obj.name)
        if obj.type == 'MESH' and obj.data.uv_textures.active:
            name += "_" + clean_name(obj.data.uv_textures.active.name)
        return name

    def _face_uv_iter(self, context, mesh):
        uv_layer = mesh.uv_layers.active.data
        polys = mesh.polygons

        if not self.export_all:
            uv_tex = mesh.uv_textures.active.data
            local_image = Ellipsis

            if context.tool_settings.show_uv_local_view:
                space_data = self._space_image(context)
                if space_data:
                    local_image = space_data.image

            for i, p in enumerate(polys):
                # context checks
                if polys[i].select and local_image in {Ellipsis,
                                                       uv_tex[i].image}:
                    start = p.loop_start
                    end = start + p.loop_total
                    uvs = tuple((uv.uv[0], uv.uv[1])
                                for uv in uv_layer[start:end])
                    #~ uv = uv_elem.uv
                    #~ if False not in uv_elem.select_uv[:len(uv)]:
                    #~     yield (i, uv)

                    # just write what we see.
                    yield (i, uvs)
        else:
            # all, simple
            for i, p in enumerate(polys):
                start = p.loop_start
                end = start + p.loop_total
                uvs = tuple((uv.uv[0], uv.uv[1]) for uv in uv_layer[start:end])
                yield (i, uvs)
                
    def _store_current_parameters(self, context):
        """Saves selected properties of this object in the scene data  
            Arguments:
            @context (Context): context received from Blender caller
        """
        params = dict()
        for aname in self._stored_attributes:
            params[aname] = getattr(self,aname)
            if _DEBUG > 1: print("stored:", aname, params[aname])
        #Store the data of this add-on in the current scene:
        context.scene[THIS_ADDON] = params
        
    def _restore_last_parameters(self, context):
        """Retrieves eventual stored properties of this object from the scene data  
            Arguments:
            @context (Context): context received from Blender caller
        """
        params = context.scene.get(THIS_ADDON,None)
        if params:
            for aname in self._stored_attributes:
                setattr(self,aname,params.get(aname,getattr(self,aname)))
                if _DEBUG > 1: print("restored:", aname, params.get(aname,"(skipped)"))
                
    def _svg_item_name(self,obj):
        """Returns name for the SVG group that will contain the mesh layout    
            Arguments:
            @obj (Object): object, which mesh will be exported
            
            When this function returns None, the obj should not be processed
        """
        if obj.type == 'MESH': #among multiple selected objects there can be some non-mesh ones!
            if self.modified:
                name = escape(obj.name) #for the meshes with applied modifiers use the object names 
            else:
                name = escape(obj.data.name) #for all the other cases use the mesh names
                
            if obj.data.uv_textures.active: 
                return "%s:%s" % (name, escape(obj.data.uv_textures.active.name))
            else:
                return None #no UV data in this mesh!  
        else:
            return None
                    
    def _write_mesh(self, fw, context, obj, name):
        """Helper function, to used process a single object  
            Arguments:
            @context (Context): context received from Blender caller
            @obj (Object): mesh object to be processed (it should contain the UV data)
            @name (str): name, which will receive its layout in the SVG file
            
           Created just to clarify the main code
        """
        if self.modified:
            mesh = obj.to_mesh(context.scene, True, 'PREVIEW')
        else:
            mesh = obj.data
            
        if mesh.uv_textures.active:
            write_mesh(fw, mesh, name, self.size[0], self.size[1], 
                                lambda: self._face_uv_iter(context, mesh),
                                self.fill_color, self.opacity, 
                                self.stroke_color, self.stroke_width,
                                self.override_color)
            
        if self.modified:
            bpy.data.meshes.remove(mesh)
        
    
    #---### Blender API Operator interface implementation
    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return (obj and obj.type == 'MESH' and obj.data.uv_textures)

    def execute(self, context):
        obj = context.active_object
        is_editmode = (obj.mode == 'EDIT')
        if is_editmode:
            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        mode = self.mode
        filepath = self.filepath
        if self.use_object_name:
            filepath = join(split(filepath)[0],self._object_filename(obj))
        filepath = ensure_ext(filepath, "." + mode.lower())
        
        file = open(filepath, "w")
        fw = file.write
        
        write_header(fw, self.size[0], self.size[1])
        if self.add_border: write_border(fw, self.size[0], self.size[1], self.stroke_width, self.stroke_color)
        
        if self.export_all and self.export_selected:
            processed = [] #helper list for the processed items (to not duplicate the SVG group id!)
            for obj in context.selected_objects:
                item_name = self._svg_item_name(obj) #it returns None, when object does not contain UV data
                if item_name and item_name not in processed:
                    self._write_mesh(fw, context, obj, item_name)
                    processed.append(item_name)
        else:
            self._write_mesh(fw, context, obj, self._svg_item_name(obj))
            
        write_footer(fw)
        
        if is_editmode:
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        file.close()
        self._store_current_parameters(context)
        return {'FINISHED'}

    def check(self, context):
        filepath = ensure_ext(self.filepath, "." + self.mode.lower())
        if filepath != self.filepath:
            self.filepath = filepath
            return True
        else:
            return False

    def invoke(self, context, event):
        self.size = self._image_size(context)
        self.filepath = self._default_filepath()
        self._restore_last_parameters(context)
        wm = context.window_manager
        wm.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row = row.split(0.35)
        row.prop(self, "export_all")
        row = row.split(1.0)
        row.enabled = self.export_all
        row.prop(self,"export_selected")
        layout.prop(self,"modified")
        layout.prop(self, "use_object_name")
        layout.prop(self, "size")
        layout.label("Faces:")
        row = layout.row()
        sub = row.split(0.48)
        sub.prop(self,"override_color",text="Fill Color:")
        sub = sub.split(1.0)
        sub.enabled = self.override_color
        sub.prop(self,"fill_color",text="")
        layout.prop(self, "opacity")
        layout.label("Edges:")
        layout.prop(self,"stroke_color", text="Line Color")
        layout.prop(self,"stroke_width", text="Line Width")
        layout.separator()
        layout.prop(self,"add_border")
        layout.separator()

def menu_func(self, context):
    self.layout.operator(ExportUVLayoutToSVG.bl_idname)


def register():
    register_module(__name__)
    bpy.types.IMAGE_MT_uvs.append(menu_func)
    if _DEBUG: print("io_mesh_uv_svg - registered!")


def unregister():
    unregister_module(__name__)
    bpy.types.IMAGE_MT_uvs.remove(menu_func)
    if _DEBUG: print("io_mesh_uv_svg - unregistered!")
    
#---##### SVG source text generation
def svg_color(color):
    """Returns the SVG expression that represents given color
        Arguments:
        @color (Color): a list or iterator containing three items: R (0), G (1), B (2)
    """
    return 'rgb(%d, %d, %d)' % tuple(int(c * 255) for c in color)
    
def write_header(fw,image_width, image_height):
    """Writes the required header of the the SVG image source text 
        Arguments:
        @fw (function): function that should be used to write the SVG text
        @image_width (int): the width of the SVG image (from add-on "Image Size" parameter)
        @image_height(int): the height of the SVG image (from add-on "Image Size" parameter)
    """
    # for making an XML compatible string
    fw('<?xml version="1.0" standalone="no"?>\n')
    fw('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" \n')
    fw('  "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n')
    fw('<svg width="%dpx" height="%dpx" viewBox="0px 0px %dpx %dpx"\n' %
       (image_width, image_height, image_width, image_height))
    #these two additional namespaces are useful for Inkscape:
    fw('     xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"\n')
    fw('     xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"\n')
    fw('     xmlns="http://www.w3.org/2000/svg" version="1.1">\n')
    desc = ("%r, (Blender %s)" % (basename(bpy.data.filepath), version_string))
    fw('<desc>%s</desc>\n' % escape(desc))
    
def write_border(fw, image_width, image_height, stroke_width, stroke_color):
    """Creates a border around the SVG image   
        Arguments:
        @fw (function): function that should be used to write the SVG text
        @image_width (int): the width of the SVG image (from add-on "Image Size" parameter)
        @image_height(int): the height of the SVG image (from add-on "Image Size" parameter)
        @stroke_width (float): the width of the outline(from add-on "Line Width" parameter)
        @stroke_color (Color): the color of the polygon's outline (from add-on "Line Color" field)
    """
    #the SVG rectangle has nominal width and height in the middle of its contour width
    #so we have to decrease it by half of the line width from each side:
    x = stroke_width/2.0
    y = stroke_width/2.0
    width = image_width - stroke_width
    height = image_width - stroke_width
    fw('<rect x="%.3f" y="%.3f" width="%.3f" height="%.3f"\n' % (x,y,width,height))
    fw('style="fill:none;stroke:%s;stroke-width:%.2g;"\n/>\n' % (svg_color(stroke_color), stroke_width))
        
def write_mesh(fw, mesh, group_name, image_width, image_height, face_iter_func, 
               fill_color,fill_opacity,stroke_color,stroke_width,override_fill_color):
    """Writes the active UV Layout of given mesh as the SVG group of polygons   
        Arguments:
        @fw (function): function that should be used to write the SVG text
        @mesh (Mesh): mesh, which active UV Layout has to be written
        @group_name (str): the id for this SVG group
        @image_width (int): the width of the SVG image (from add-on "Image Size" parameter)
        @image_height(int): the height of the SVG image (from add-on "Image Size" parameter)
        @face_iter_func (function): function that iterates through the UV faces
        @fill_color (Color): the color of the faces (from add-on "Fill Color" parameter value)
        @fill_opacity (float): the opacity of the faces (from add-on "Fill Opacity" parameter)
        @stroke_color (Color): the color of the polygon's outline (from add-on "Line Color" field)
        @stroke_width (float): the width of the outline(from add-on "Line Width" parameter)
        @override_fill_color (bool): when False, use the material color instead of the @fill_color
                                     (it is the state of the add-on "Fill Color" check box)
    """
    
    group_style = "fill:%s;fill-opacity:%.2g;stroke:%s;stroke-opacity:1;stroke-width:%.2g;"  
    group_style %= (svg_color(fill_color),fill_opacity,svg_color(stroke_color),stroke_width)
    fw('<g id="%s"\n style="%s">\n' % (group_name, group_style))

    fill_settings = []
    for mat in mesh.materials if mesh.materials else [None]:
        if mat:
            fill_settings.append(svg_color(mat.diffuse_color))
        else:
            fill_settings.append(None)

    faces = mesh.polygons
    for i, uvs in face_iter_func():
        try:  # rare cases material index is invalid.
            fill = fill_settings[faces[i].material_index]
        except IndexError:
            fill = None

        fw('<polygon')
        if fill and not override_fill_color: 
            fw(' fill="%s" ' % fill)

        fw(' points="')

        for j,uv in enumerate(uvs):
            x, y = uv[0], 1.0 - uv[1]
            fw('%.3f,%.3f ' % (x * image_width, y * image_height))
            
        fw('" />\n')
        
    fw('</g>\n')
    
def write_footer(fw):
    """Writes the required footer of the the SVG source text   
        Arguments:
        @fw (function): function that should be used to write the SVG text
    """
    fw('\n')
    fw('</svg>\n')
    
#---##### Main code
if __name__ == "__main__":
    register()
