# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
bl_info = {
    "name": "Add Empty Parent",
    "author": "Witold Jaworski",
    "version": (1, 0, 0),
    "blender": (2, 6, 6),
    "location": "View 3D > Specials [W-key] > Add Empty Parent",
    "category": "Object",
    "description": "Adds an empty object as the parent of the active object",
    "warning": "",
    "wiki_url": "",
    "tracker_url": ""
    }
#----- updates
#----- imports
import bpy
from bpy.utils import register_module, unregister_module
####constants---------------------------------------------------------------
DEBUG = 1 # Debug level. When nonzero, some diagnostics texts are placed at 
        #   output, during processing. Level = 1 generates some genral
        #   statements, Level = 5 - intermediate, while Level = 9 is the most 
        #   detailed
####globals-----------------------------------------------------------------

####classes-----------------------------------------------------------------
####module functions--------------------------------------------------------
def select(context, objects):
    """Helper function: selects specified objects
       Arguments:
            @context (Context):    current context 
            @objects (list):      list of the objects to be selected (the first element will become the active one) 
       This function should be called in the Object Mode, only!
    """    
    bpy.ops.object.select_all(action='DESELECT') #clear teh current selection
    for obj in objects: #select these objects
        obj.select = True
    if objects: #mark the first on the list as the active one
        bpy.context.scene.objects.active = objects[0]     

def parent_set(context,parent,child):
    """Helper function: sets the parent
       Arguments:
            @context (Context):    current context
            @parent (Object):     the parent object
            @child (Object):      the child object 
       This function should be called in the Object Mode, only!
    """
    select(context, [parent, child])    
    bpy.ops.object.parent_set(type='OBJECT', xmirror=False, keep_transform=True) 

def parent_clear(context,child):
    """Helper function: clears the parent
       Arguments:
            @context (Context):    current context
            @child (Object):      a child object 
       This function should be called in the Object Mode, only!
    """
    select(context, [child])    
    bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
           
#--- ### Blender Operators
class AddEmptyParent(bpy.types.Operator):
    bl_idname = "object.add_empty_parent"
    bl_label = "Add Empty Parent"
    bl_description = "Add an empty object as the parent of the active object and a child of its parent"
    #--- parameters 
    #--- Blender interface methods
    @classmethod
    def poll(cls,context):
        return (context.mode == 'OBJECT')
       
    def invoke(self, context, event):
        #similar check, like in IntersectMeshes, but different way of communication:
        obj = context.object
        
        if not obj:
            msg = "No active object"
            self.report(type={'ERROR'}, message=msg)
            return {'CANCELLED'}
        else:
            return self.execute(context)
    
    def execute(self, context):
        #bpy.ops.object.editmode_toggle()
        obj = context.object
        parent = obj.parent
        parent_clear(context, obj)
        rotation = list(obj.rotation_euler)
        #create an empty object (on the current layer)
        bpy.ops.object.empty_add(type='ARROWS', view_align=False, location=obj.location, rotation=(0, 0, 0), layers=list(obj.layers))
        empty = context.object
        empty.name = obj.name + ".Parent"
        obj.rotation_euler = [0,0,0]
        parent_set(context,empty, obj)
        empty.rotation_euler = rotation
        if parent:
            parent_set(context,parent,empty)
        select(context,[empty])
        return {'FINISHED'}
        
def menu_object_draw(self, context):
    if (context.object):
        self.layout.operator_context = 'INVOKE_REGION_WIN'
        self.layout.operator(AddEmptyParent.bl_idname)

#--- ### Register
def register():
    register_module(__name__)
    #bpy.types.VIEW3D_MT_edit_mesh_edges.prepend(menu_draw)
    #bpy.types.VIEW3D_MT_object.prepend(menu_draw)
    bpy.types.VIEW3D_MT_object_specials.prepend(menu_object_draw)
    
def unregister():
    bpy.types.VIEW3D_MT_object_specials.remove(menu_object_draw)
    #bpy.types.VIEW3D_MT_edit_mesh_edges.remove(menu_draw)
    #bpy.types.VIEW3D_MT_object.remove(menu_draw)
    unregister_module(__name__)
    
if DEBUG > 0:
    print("object_add_parent.py: new version loaded!") #just to be sure, that we are testing what we should
    
if __name__ == '__main__':
    register()
